import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import colorsys
import fund_alchemy.general_tools as gt
from fund_alchemy.config import matplotlib_custom_style


def use_style():
    plt.style.use(matplotlib_custom_style)


def comprehensive_analysis(series, rf=0.03):
    """计算基于净值序列的分析指标.

    parameters:
    ----------
    series -- 净值序列
    rf -- 无风险利率

    returns:
    dict, 各指标计算结果
    """

    annual_return, annual_volatility, max_drawdown, \
        sharpe_ratio, sortino_ratio, calmar_ratio \
        = gt.Calculate_Evaluation_Indicators(series, rf, ALL_Indicator=True)
    burke_ratio = gt.Calculate_Modified_Burke_Ratio(series, rf)
    proportion_of_high_days = gt.Calculate_days_for_highs(series)
    min_holding_weeks1 = gt.Calculate_days_for_target_win_rate(
        series, min_holding_days=4, target_win_rate=0.8)
    min_holding_weeks2 = gt.Calculate_days_for_target_win_rate(
        series, min_holding_days=4, target_win_rate=0.9)
    Var, CVar = gt.Calculate_Var(series)
    maxdrawdown_initial_date, reach_maxdrawdown_days, maxdrawdown_date,\
        recovery_days, recovery_date = gt.Max_Drawdown_Details(series)
    N, win_ratio, PnL_ratio = gt.calculate_PnL_ratio(series)

    return {
        '年化收益率': annual_return,
        '年化波动率': annual_volatility,
        '最大回撤': max_drawdown,
        '夏普比率': sharpe_ratio,
        '索提诺比率': sortino_ratio,
        '卡玛比率': calmar_ratio,
        '伯克比率': burke_ratio,
        '创新高周数占比': proportion_of_high_days,
        '八成胜率最小持有周数': min_holding_weeks1,
        '九成胜率最小持有周数': min_holding_weeks2,
        'Var-95%': Var,
        'CVar-95%': CVar,
        '最大回撤前高日期': maxdrawdown_initial_date,
        '最大回撤形成天数': reach_maxdrawdown_days,
        '最大回撤日期': maxdrawdown_date,
        '最大回撤恢复天数': recovery_days,
        '最大回撤恢复日期': recovery_date,
        '净值数目': N+1,
        '胜率(r>0)': win_ratio,
        '盈亏比(r>0/r<0)': PnL_ratio,
    }


def draw_overview(df, fre="w", years=["2020", '2021', "2022"],
                  color_pct_clip=(-5, 5), show_excess_return=False):
    """净值概览图.

    parameters:
    -----------
    df -- pd.DataFrame
        两列净值序列，index为datetime.date，第一列为产品净值，第二列为业绩基准
    fre -- str, "m/w/d"
        数据频率
    years -- list
        月度收益显示的年份
    color_pct_clip -- tuple
        颜色最深对应的涨跌幅上下限
    show_excess_return -- bool
        是否展示超额收益

    returns:
    ----------
    生成matplotlib中的figure
    """

    fre_dict = {'w': 12, "d": 63}

    benchmark = df.columns[1]
    name = df.columns[0]
    df = df.fillna(method='ffill').fillna(method='bfill')
    df[[name, benchmark]] /= df[[name, benchmark]].iloc[0, :]
    df['alpha'] = ((df[name].pct_change() - df[benchmark].pct_change()).fillna(0) + 1).cumprod()
    df['dd'] = df[name]/df[name].expanding().max() - 1
    df['v'] = (df[name].pct_change().dropna()
               .rolling(fre_dict[fre]).std()*(52)**(0.5)*100).round(decimals=2)
    df['v_bk'] = (df[benchmark].pct_change().dropna()
                  .rolling(fre_dict[fre]).std()*(52)**(0.5)*100).round(decimals=2)

    res = comprehensive_analysis(df[name])
    res_bk = comprehensive_analysis(df[benchmark])

    fig = plt.figure(figsize=(9.0551181, 4.5275591*1.25))
    gs = GridSpec(10, 4, figure=fig)
    ax1 = fig.add_subplot(gs[:4, :-1])
    ax2 = fig.add_subplot(gs[4:6, :-1], sharex=ax1)
    ax3 = fig.add_subplot(gs[6:8, :-1], sharex=ax1)
    ax4 = fig.add_subplot(gs[8:10, :-1])
    ax5 = fig.add_subplot(gs[:, -1])

    l1 = ax1.plot(df[name])
    l2 = ax1.plot(df[benchmark])
    if show_excess_return:
        ax1_twin = ax1.twinx()
        l3 = ax1_twin.plot(df['alpha'], '--', linewidth=0.5, color='black')
        ax1.legend(l1+l2+l3, [name, benchmark, '超额收益'], loc='upper left')
        ax1.spines['right'].set_visible(True)
        IR = gt.Calculate_IR(df[[name, benchmark]])
    else:
        ax1.legend(l1+l2, [name, benchmark], loc='upper left')
        IR = "NaN"

    ax2.stackplot(df.index, 100*df['dd'], color=(0.8, 0.8, 0.8, 1))
    ax2.legend(['动态回撤(%)'], loc='lower left')

    ax3.plot(df[['v', 'v_bk']])
    ax3.legend(['组合滚动3个月年化波动率', '基准滚动3个月年化波动率'], loc='upper left')

    df_monthly = gt.switch_df_date_fre(df[[name, benchmark]], fre="m")
    df_monthly = df_monthly.pct_change().iloc[1:, :]
    df_monthly['y'] = [str(i.year) for i in df_monthly.index]
    df_monthly['m'] = [str(i.month) for i in df_monthly.index]
    m1_raw = gt.kvalue2table(df_monthly, "m", name, "y")
    m2_raw = gt.kvalue2table(df_monthly, "m", benchmark, "y")

    m1 = pd.DataFrame(index=years, columns=[str(i) for i in range(1, 13)])
    for i in m1.index:
        for j in m1.columns:
            if (i in m1_raw.index) and (j in m1_raw.columns):
                m1.loc[i, j] = m1_raw.loc[i, j]
            else:
                m1.loc[i, j] = float("nan")

    m2 = pd.DataFrame(index=years, columns=[str(i) for i in range(1, 13)])
    for i in m2.index:
        for j in m2.columns:
            if (i in m2_raw.index) and (j in m2_raw.columns):
                m2.loc[i, j] = m2_raw.loc[i, j]
            else:
                m2.loc[i, j] = float("nan")

    m = pd.DataFrame()
    for i in range(len(m1)):
        m = pd.concat([m, m1.iloc[[i], :]])
        m = pd.concat([m, m2.iloc[[i], :]])

    m = m * 100
    m = m.astype(float).round(2)

    def color_transfer(value):
        if value == value:
            U = (1 - value) * 100
            S = 255
            L = 180
            rgb = colorsys.hls_to_rgb(U/255, L/255, S/255)
            rgb = tuple([round(i, 4) for i in rgb])
            return rgb
        else:
            return (1, 1, 1)

    colormap = m.clip(color_pct_clip[0], color_pct_clip[1]).copy()
    colormap = colormap.applymap(
        lambda x: x / (color_pct_clip[1] - color_pct_clip[0]) + 0.5
        ).values.tolist()
    for i in range(len(colormap)):
        for j in range(len(colormap[i])):
            colormap[i][j] = color_transfer(colormap[i][j])

    m = m.applymap(lambda x: str(x)+"%").replace("nan%", "").values.tolist()
    rowLabels = ["%s-%s" % (x, y) for y in years for x in ['策略', '基准']]
    ax4.table(
        cellText=m,
        rowLabels=rowLabels,
        rowLoc='center',
        colLabels=["%s月" % i for i in range(1, 13)],
        colWidths=[0.077 for i in range(12)],
        cellColours=colormap,
        cellLoc='center',
        loc='upper right')
    ax4.set_axis_off()

    colLabels = ['基金表现', '比较基准']
    rowLabels = list(res)

    cellText = [['%.2f%%' % (100*res['年化收益率']), '%.2f%%' % (100*res_bk['年化收益率'])],
                ['%.2f%%' % (100*res['年化波动率']), '%.2f%%' % (100*res_bk['年化波动率'])],
                ['%.2f%%' % (100*res['最大回撤']), '%.2f%%' % (100*res_bk['最大回撤'])],
                ['%.2f' % res['夏普比率'], '%.2f' % res_bk['夏普比率']],
                ['%.2f' % res['索提诺比率'], '%.2f' % res_bk['索提诺比率']],
                ['%.2f' % res['卡玛比率'], '%.2f' % res_bk['卡玛比率']],
                ['%.2f' % res['伯克比率'], '%.2f' % res_bk['伯克比率']],
                ['%.2f%%' % res['创新高周数占比'], '%.2f%%' % res_bk['创新高周数占比']],
                ['%.0f' % res['八成胜率最小持有周数'], '%.0f' % res_bk['八成胜率最小持有周数']],
                ['%.0f' % res['九成胜率最小持有周数'], '%.0f' % res_bk['九成胜率最小持有周数']],
                ['%.2f%%' % (100*res['Var-95%']), '%.2f%%' % (100*res_bk['Var-95%'])],
                ['%.2f%%' % (100*res['CVar-95%']), '%.2f%%' % (100*res_bk['CVar-95%'])],
                ['%s' % res['最大回撤前高日期'], '%s' % res_bk['最大回撤前高日期']],
                ['%.0f' % res['最大回撤形成天数'], '%.0f' % res_bk['最大回撤形成天数']],
                ['%s' % res['最大回撤日期'], '%s' % res_bk['最大回撤日期']],
                ['%.0f' % res['最大回撤恢复天数'], '%.0f' % res_bk['最大回撤恢复天数']],
                ['%s' % res['最大回撤恢复日期'], '%s' % res_bk['最大回撤恢复日期']],
                ['%s' % res['净值数目'], '%s' % res_bk['净值数目']],
                ['%s' % res['胜率(r>0)'], '%s' % res_bk['胜率(r>0)']],
                ['%s' % res['盈亏比(r>0/r<0)'], '%s' % res_bk['盈亏比(r>0/r<0)']],
                ]

    if show_excess_return:
        rowLabels = rowLabels + ["信息比率"]
        cellText = cellText + [['%s' % IR, 'NaN']]

    ax5.table(
        cellText=cellText,
        rowLabels=rowLabels,
        rowLoc='center',
        colLabels=colLabels,
        colWidths=[0.8, 0.8],
        cellLoc='center',
        loc='upper left')
    ax5.set_axis_off()

    plt.subplots_adjust(left=0.06, right=0.9, top=0.9, bottom=0.05,
                        wspace=1, hspace=0.15)

    return fig


if __name__ == '__main__':
    pass



import pandas as pd
import numpy as np
import datetime as dt

# ==============================
# 常用自定义函数集
# Author: xxf
# 更新日期: 2023-07-28

# 函数目录:
# 1.kvalue2table -- 转换数据格式
# 2.ApplyFunc -- 替换变量
# 3.CutList -- 切分列表
# 4.d -- 将字符串日期格式转换为datetime.date
# 5.switcher -- 字符串与列表的互相转换
# 6.switch_df_date_fre -- 改变“日期”列频率
# 7.Calculate_Evaluation_Indicators -- 计算序列业绩
# 8.Calculate_days_for_target_win_rate -- 计算目标胜率最小持有天数
# 9.Calculate_IR -- 计算信息比率
# 10.GetRandomSample -- 有放回抽样
# 11.GetPercentile -- 返回特定值在样本中的分位数
# 12.MWRR -- 计算Money-weighted rate of return
# 13.get_rpt_date -- 获取目标日期最近的财报日(假定2,5,8,11披露)
# 14.get_current_rpt -- 获取目标日期当前所处的季度
# 15.get_last_month_end_date -- 获取目标日期最近一个月最后一天日期
# 16.Calculate_Modified_Burke_Ratio -- 计算修正伯克比率
# 17.Calculate_Var -- 计算VaR和CVaR
# 18.Max_Drawdown_Details -- 计算最大回撤细节
# 19.calculate_drawdown_max_days -- 计算水下最长天数
# 20.calculate_PnL_ratio -- 计算胜率和盈亏比
# 21.str_clearing -- 文本清洗
# 22.ym_ago -- 获取目标日期之前n个月或年的日期
# 23.detect_fre -- 检测序列的频率
# 24.drop_unstart -- 剔除产品净值未变动的区间
# 25.interpolate -- 直线内插法
# 26.find_date -- 从字符串中识别日期
# 27.table2value -- 转换数据格式
# ==============================


def kvalue2table(df, label='基金代码', variable=None, index='日期'):
    """用于转换key-value形式的数据至DataFrame.

    parameters:
    -----------
    df -- DataFrame
    label -- 分类列名
    variable -- 变量列名

    returns:
    --------
    pd.DataFrame
    """

    if variable is None:
        variable = df.columns[-1]
    CODEs = list(df[label].drop_duplicates())
    DATEs = sorted(list(df[index].drop_duplicates()))
    data = pd.DataFrame(index=DATEs)
    for code in CODEs:
        series = df[df[label] == code].set_index(index)[variable]
        series.name = code
        data = data.join(series)

    return data


def ApplyFunc(x, DICT, trans=None, fillna=float('nan')):
    """用于转换DataFrame中的数据, 配合apply函数使用.

    parameters:
    -----------
    x -- 需转换的变量.
    DICT -- dict, 对应的转换关系
    trans -- dict, 对结果的第二次转换
    fillna -- 缺省值填充值

    returns:
    --------
    value
    """

    try:
        if trans is None:
            return DICT[x]
        else:
            return trans[DICT[x]]
    except(Exception):
        return fillna


def CutList(L, n=500):
    '''将大列表L切分成每份数量为n的小列表们.

    parameters:
    -----------
    L -- list, 母列表
    n -- int, 字列表元素个数

    returns:
    --------
    list
    '''

    res = []
    for i in range(int(len(L)/n)+1):
        if len(L[i*n:(i+1)*n]) != 0:
            res.append(L[i*n:(i+1)*n])
    return res


def d(date, t='date'):
    '''将str的日期转化为dt.date或dt.datetime

    parameters:
    -----------
    date -- str, 日期
    t -- date/datetime

    returns:
    --------
    dt.date / dt.datetime
    '''

    if t == 'date':
        if '-' in date:
            return dt.datetime.strptime(date, "%Y-%m-%d").date()
        else:
            return dt.datetime.strptime(date, "%Y%m%d").date()
    elif t == 'datetime':
        if '-' in date:
            return dt.datetime.strptime(date, "%Y-%m-%d")
        else:
            return dt.datetime.strptime(date, "%Y%m%d")


def switcher(x, label=';'):
    """实现list与str的互相转换.

    Parameters:
    ----------
    x -- list/str
        列表或字符串
    label -- str
        分割符号

    returns:
    --------
    list / str
    """

    if type(x) == str:
        return x.split(label)
    elif type(x) == list:
        res = ''
        for i in range(len(x)):
            res += str(x[i])
            if i != len(x) - 1:
                res += label
        return res
    elif x != x:
        return []
    else:
        return x


def switch_df_date_fre(df, fre='w', date_index=True):
    '''转换日频净值数据为周/月频.

    parameters:
    -----------
    df -- pd.DataFrame, 净值数据
    fre -- 'w'/'m', 周频/月频
    date_index -- df中index是否为日期, 若否则需含一列命名为“日期”的序列.

    returns:
    --------
    df -- pd.DataFrame
    '''

    df = df.copy()
    shape = len(df.shape)

    if date_index is True:
        old_name = df.index.name
        df.index.name = '日期'
        df = df.reset_index()

    if fre == 'w':
        df['_slice'] = df['日期'].apply(lambda x: x.strftime("%Y%W"))
        first = df.iloc[0:1, :]
        last = df.iloc[-1:, :]
        mid = df.drop_duplicates(subset=['_slice'], keep='last')
        df = first.append(mid).append(last).drop_duplicates()
        df = df.drop(['_slice'], axis=1).reset_index(drop=True)
    elif fre == 'm':
        df['_slice'] = df['日期'].apply(lambda x: x.strftime("%Y%m"))
        first = df.iloc[0:1, :]
        last = df.iloc[-1:, :]
        mid = df.drop_duplicates(subset=['_slice'], keep='last')
        df = first.append(mid).append(last).drop_duplicates()
        df = df.drop(['_slice'], axis=1).reset_index(drop=True)

    if date_index:
        df = df.set_index('日期')
        df.index.name = old_name

    if shape == 2:
        return df.copy()
    else:
        return df[df.columns[0]].copy()


def Calculate_Evaluation_Indicators(series, rf=0.03, ALL_Indicator=False):
    '''函数用来计算年化收益率、年化波动率、最大回撤、夏普、索提诺、卡玛。

    parameters:
    -----------
    series -- pd.Series, index为dt.date的净值序列.
    rf -- 无风险利率.
    ALL_Indicator -- True/False, 是否添加索提诺和卡玛.

    returns:
    -------
    4或6个指标的tuple
    '''

    if len(series) < 5:
        if not ALL_Indicator:
            return 0, 0, 0, 0
        else:
            return 0, 0, 0, 0, 0, 0
    series = series.dropna().astype(float)
    series_original = series.copy()
    series = series.reset_index()
    series.columns = ['日期', '净值']
    series['_seq'] = series['日期'].apply(lambda x: int(x.strftime('%w')))
    series['_choose'] = False
    series.loc[0, '_choose'] = True
    for r in range(1, series.shape[0]-1):
        if series.loc[r, '_seq'] >= series.loc[r+1, '_seq']:
            series.loc[r, '_choose'] = True
    series.loc[series.shape[0]-1, '_choose'] = True
    series = series[series['_choose']]\
        .drop(['_seq', '_choose'], axis=1)\
        .set_index('日期')['净值']
    rr = round((
        (series.iloc[-1]/series.iloc[0])
        ** (365./(series.index[-1]-series.index[0]).days)
        - 1.), 4)
    std = round((series.pct_change().std()*np.sqrt(52.)), 4)
    if std != 0:
        sharpe = round((rr-rf)/std, 4)
    else:
        sharpe = 0
    maxdrawdown = round((
        1.-(series_original/series_original.expanding().max())
        ).max(), 4)
    if not ALL_Indicator:
        return rr, std, maxdrawdown, sharpe
    else:
        d_std = series.pct_change().dropna()\
            .apply(lambda x: min(x, 0)).std()*np.sqrt(52.)
        if d_std != 0:
            sortino = round((rr-rf)/d_std, 4)
        else:
            sortino = 0
        if maxdrawdown != 0:
            calmar = round(rr / maxdrawdown, 4)
        else:
            calmar = 0
        return rr, std, maxdrawdown, sharpe, sortino, calmar


def Calculate_days_for_target_win_rate(series, min_holding_days=20,
                                       target_win_rate=0.8, **kwargs):
    """用于计算X成胜率(区间收益大于0)所需要的最小持有天数.

    parameters:
    -----------
    series -- pd.Series, index为dt.date的净值序列.
    min_holding_days -- int, 最少持有天数
    target_win_rate -- float, 目标胜率.
    kwargs:
        ShowDistribution -- int, 持有天数

    return:
    -------
    day -- 最小持有天数
    series -- 滚动区间收益序列
    """

    df = series.copy()
    df = df.reset_index()
    df.columns = ['日期', '净值']
    rolling_return = {}

    day = 0
    for p in range(min_holding_days, len(df)-min_holding_days):
        rolling_return[p] = df['净值'] / df['净值'].shift(p) - 1
        if (rolling_return[p] > 0).map({True: 1, False: 0}).sum() \
                > target_win_rate*(len(df)-p):
            day = p
            break

        # df[p] = df['净值'] / df['净值'].shift(p) - 1
        # if (df[p] > 0).map({True: 1, False: 0}).sum() \
        #         > target_win_rate*(len(df)-p):
        #     day = p
        #     break

    if 'ShowDistribution' in kwargs:
        newframe = pd.DataFrame(rolling_return)
        print(newframe)
        df = df.merge(newframe, left_index=True, right_index=True)
        return (day, df[kwargs['ShowDistribution']])

    return day


def Calculate_days_for_highs(series):
    """用于计算创新高天数占总天数的比例.

    parameters:
    -----------
    series -- pd.Series, index为dt.date的净值序列.

    return:
    -------
    pct -- float, 创高新天数占比
    """

    df = series.copy()
    df = df.reset_index()
    df.columns = ['日期', '净值']

    days = len(df)-1
    df['new_high'] = (
        df['净值'] >= df['净值'].expanding().max()
        ).replace({True: 1, False: 0})
    pct = round(100*df['new_high'].sum() / days, 2)

    return pct


def Calculate_IR(df):
    """用于计算信息比率.

    parameters:
    -----------
    df -- pd.DataFrame, 包含两列的表, index为日期, 第二列为净值序列, 第三列为基准序列.

    return:
    -------
    IR -- float, 信息比率
    """

    df = df.copy()
    df.columns = ['净值', '基准']
    df = df.astype(float).pct_change(fill_method=None)
    df['active_return'] = df['净值'] - df['基准']
    IR = df['active_return'].mean() / df['active_return'].std()

    return round(IR, 4)


def GetRandomSample(sample):
    """有放回抽样.

    parameters:
    -----------
    sample -- list, 样本

    return:
    -------
    RES -- list, 有放回随机抽样样本
    """

    N = len(sample)
    rdm = np.random.random(N)*N
    RES = [sample[int(i)] for i in rdm]
    return RES


def GetPercentile(x, sample):
    """返回特定值在样本中的分位数.

    parameters:
    -----------
    x -- 特定值
    sample -- list, 样本

    return:
    -------
    pct -- float, 分位数
    """

    sample = sorted(sample)
    for i in range(len(sample)):
        if x >= sample[i]:
            pass
        else:
            return float(i)/len(sample)
        if i == len(sample) - 1:
            return 1


def MWRR(df, start_ret=-1, end_ret=1, error=0.00001):
    """计算Money-weighted rate of return, 适用场景如基金定投的收益计算.

    parameters:
    -----------
    cash_flows -- pd.Series, index为日期, 值为现金流数值
    end_date -- dt.datetime, 期末日期
    end_asset -- float, 期末资产
    start_ret -- float, 年化收益率测试起始值
    end_ret -- float, 年化收益率测试终止值
    error -- float, 计算精度

    return:
    -------
    float, MWRR
    """

    df = df.reset_index()
    df.columns = ['date', 'nav', 'inflow']
    df.loc[0, 'shares'] = df.loc[0, 'inflow'] / df.loc[0, 'nav']
    for r in df.index[1:]:
        df.loc[r, 'shares'] = \
            df.loc[r, 'inflow'] / df.loc[r, 'nav'] + df.loc[r - 1, 'shares']
    df['mv'] = df['nav'] * df['shares']
    cash_flows = df.set_index('date')['inflow']
    end_date = df['date'].iloc[-1]
    end_asset = df['mv'].iloc[-1]

    def f(r0, r1):
        diff0 = 0
        ret0 = 0
        for ret in np.linspace(r0, r1, 3):
            SUM = 0
            for i in range(len(cash_flows)):
                cash = cash_flows.iloc[i]
                date = cash_flows.index[i]
                d = (end_date - date).days
                SUM += cash*(1+ret)**(d/365)
            diff = SUM - end_asset
            if diff * diff0 < 0:
                return ret0, ret
            else:
                diff0 = diff
                ret0 = ret
        return None

    check_range = f(start_ret, end_ret)
    if check_range is None:
        print('Out of range!')
        return
    else:
        a, b = check_range
        while 1:
            if b - a <= error:
                return round((a+b)/2, 4), end_asset
            else:
                a, b = f(a, b)


def get_rpt_date(date, fisicalyear=False, str_output=True):
    """获取目标日期最近的财报日(假定2,5,8,11披露).

    parameters:
    -----------
    date -- str, 日期
    fisicalyear -- bool, 是否获取财年
    str_output -- bool, 是否返回str日期

    return:
    -------
    str, 财报日期
    """

    if type(date) == str:
        y = date[:4]
        date = d(date)
    else:
        y = str(date.year)

    if fisicalyear:
        if date < d(y + '0331'):
            output = str(int(y)-2) + '1231'
        else:
            output = str(int(y)-1) + '1231'
    else:
        if date < d(y + '0201'):
            output = str(int(y)-1)+'0930'
        elif date < d(y + '0501'):
            output = str(int(y)-1)+'1231'
        elif date < d(y + '0801'):
            output = y + '0331'
        elif date < d(y + '1101'):
            output = y + '0630'
        else:
            output = y + '0930'

    if str_output:
        return output
    else:
        return d(output)


def get_current_rpt(date, str_output=False):
    """获取目标日期当前所处的季度.

    parameters:
    -----------
    date -- str, 日期
    str_output -- bool, 是否返回str日期

    return:
    -------
    str / dt.date
    """

    if type(date) == str:
        y = date[:4]
        date = d(date)
    else:
        y = str(date.year)

    if date <= d(y + '0331'):
        output = y + '0331'
    elif date <= d(y + '0630'):
        output = y + '0630'
    elif date <= d(y + '0930'):
        output = y + '0930'
    elif date <= d(y + '1231'):
        output = y + '1231'

    if str_output:
        return output
    else:
        return d(output)


def get_last_month_end_date(date, str_output=True):
    """获取目标日期最近一个月最后一天日期.

    parameters:
    -----------
    date -- str, 日期
    str_output -- bool, 是否返回str日期

    return:
    -------
    str / dt.date
    """

    if type(date) == str:
        y = int(date[:4])
        m = int(date[4:6])
        date = d(date)
    else:
        y = date.year
        m = date.month
    output = (dt.datetime(y, m, 1) - dt.timedelta(days=1)).date()
    if str_output:
        return output.strftime('%Y%m%d')
    else:
        return output


def Calculate_Modified_Burke_Ratio(series, rf):
    """计算修正伯克比率.

    parameters:
    -----------
    series -- pd.Series, 净值序列
    rf -- float, 无风险利率

    return:
    -------
    float, round 4
    """

    if len(series) < 5:
        return float('nan')
    series = series.dropna().astype(float)
    dd = series/series.expanding().max() - 1.
    mssd = ((dd**2).sum()/len(dd))**(0.5)
    if mssd == 0:
        return 999
    rr = ((series.iloc[-1]/series.iloc[0])
          ** (365./(series.index[-1]-series.index[0]).days)
          - 1.)
    return round((rr - rf)/mssd, 4)


def Calculate_Var(series, a=0.05, pct=False):
    """计算Historical Value at Risk历史在险价值.

    parameters:
    -----------
    series -- pd.Series, 净值/收益率序列
    a -- float, 置信度
    pct -- bool, 输入的序列是涨跌还是净值

    return:
    -------
    Var, CVar
    """
    if len(series) > 20:
        if pct:
            r = series.dropna().sort_values().reset_index(drop=True)
        else:
            r = series.pct_change().dropna().sort_values().reset_index(drop=True)
        threshold = int(len(r)*a)-1
        Var = r.loc[threshold]
        CVar = r[:threshold].mean()

        return round(Var, 4), round(CVar, 4)

    else:
        return float("nan"), float("nan")


def Max_Drawdown_Details(series):
    """计算最大回撤细节.

    parameters:
    -----------
    series -- pd.Series, 净值序列

    return:
    -------
    maxdrawdown_initial_date, reach_maxdrawdown_days, maxdrawdown_date,
    recovery_days, recovery_date
    """

    drawdown = 1. - series/series.expanding().max()
    maxdrawdown = drawdown.max()
    maxdrawdown_date = drawdown[drawdown == maxdrawdown].index[-1]

    drawdown1 = drawdown[:maxdrawdown_date]
    maxdrawdown_initial_date = drawdown1[drawdown1 == 0].index[-1]
    reach_maxdrawdown_days = (maxdrawdown_date - maxdrawdown_initial_date).days

    drawdown2 = drawdown[maxdrawdown_date:]
    if (drawdown2 == 0).sum() > 0:
        recovery_date = drawdown2[drawdown2 == 0].index[0]
        recovery_days = (recovery_date - maxdrawdown_date).days
        recovery_date = recovery_date.strftime('%Y%m%d')
    else:
        recovery_date = float('nan')
        recovery_days = float('nan')

    return maxdrawdown_initial_date.strftime('%Y%m%d'), \
        reach_maxdrawdown_days, maxdrawdown_date.strftime('%Y%m%d'),\
        recovery_days, recovery_date


def calculate_drawdown_max_days(series):
    """计算水下最长天数.

    parameters:
    -----------
    series -- pd.Series, 净值序列

    return:
    -------
    drawdown_max_days, drawdown_max_days_recovery_date
    """

    drawdown = 1. - series/series.expanding().max()
    res = pd.DataFrame()
    last_high = drawdown.index[0]
    for date in drawdown.index:
        if drawdown.loc[date] == 0:
            last_high = date
            res.loc[date, 'drawdown_days'] = 0
        else:
            res.loc[date, 'drawdown_days'] = (date - last_high).days

    return int(res['drawdown_days'].max()), \
        res[res['drawdown_days'] == res['drawdown_days'].max()].index[-1]


def calculate_PnL_ratio(series, target=0):
    """计算胜率和盈亏比.

    parameters:
    -----------
    series -- pd.Series, 净值序列

    return:
    -------
    N, 收益率数量 
    win_ratio, 收益率大于target的占比
    PnL_ratio, 收益率大于target的收益与小于target的收益的比值
    """

    r = series.dropna().pct_change().dropna()
    N = len(r)
    win_ratio = r.loc[lambda x: x > target].count() / N
    PnL_ratio = -1*r.loc[lambda x: x > target].mean() \
        / r.loc[lambda x: x < target].mean()

    return N, round(win_ratio, 4), round(PnL_ratio, 4)


def str_clearing(text):
    """文本清洗.

    parameters:
    -----------
    text -- str, 文本信息

    return:
    -------
    str
    """

    def remove_items(text, symbol):
        a = text.split(symbol)
        b = ''
        for i in a:
            b += i
        return b

    text = remove_items(text, '\n')
    text = remove_items(text, '\r')
    text = remove_items(text, ' ')

    return text


def ym_ago(date, distance, Type, output="date"):
    """获取目标日期之前n个月或年的日期(日期推算).

    parameters:
    -----------
    date -- dt.date, 目标日期
    distance -- float, 距离
    Type -- "m"/"y", 月或年
    output -- "date"/"str/"datetime", 返回类型

    return:
    -------
    dt.date/str/dt.datetime
    """

    y, m, d = date.year, date.month, date.day
    if Type == 'm':
        m1 = m - distance
        if m1 <= 0:
            m1 = 12 + m1
            y1 = y - 1
            d1 = d
        else:
            m1 = m1
            y1 = y
            d1 = d
    elif Type == 'y':
        y1 = y - distance
        m1 = m
        d1 = d
    if output == 'date':
        return dt.datetime(y1, m1, d1).date()
    elif output == "str":
        return dt.datetime(y1, m1, d1).strftime("%Y%m%d")
    elif output == "datetime":
        return dt.datetime(y1, m1, d1)


def detect_fre(series):
    '''检测序列的频率.

    parameters:
    -----------
    series -- pd.Series, 以datetime.date为index的序列.

    returns:
    --------
    str, "d/w/m", 频率
    '''

    if (np.mean([i.day for i in series.index[-3:]]) >= 26) \
            and (series.index[-1].month != series.index[-2].month):
        return 'm'
    elif np.mean([i.weekday() for i in series.index[-12:]]) >= 3.5:
        return 'w'
    else:
        return 'd'


def drop_unstart(series, threshold=0.0001):
    '''剔除产品备案至实际运作这段净值未变动的区间.

    parameters:
    -----------
    series -- pd.Series, 以datetime.date为index的序列.

    returns:
    --------
    pd.Series
    '''

    if len(series) > 0:
        nav0 = series.iloc[0]
        need_to_drop = -1
        for i in series.index[1:]:
            nav_i = series.loc[i]
            need_to_drop += 1
            if abs(nav_i - nav0) <= threshold:
                pass
            else:
                break
        if need_to_drop == 0:
            pass
        else:
            series = series[need_to_drop-1:]

    return series


def interpolate(v0, v1, N):
    """直线内插法.

    parameters:
    ----------
    v0 -- 起始值
    v1 -- 终止值
    N -- 点位数目

    return:
    ----------
    list
    """

    d = (v1 - v0)/(N+1)
    s = [v0]
    for i in range(N):
        s.append(v0+(i+1)*d)
    s.append(v1)

    return s


def find_date(string):
    """从字符串中识别日期.

    Parameters:
    ----------
    string -- str, 字符串

    Return:
    ----------
    datetime.date or None, 日期
    """

    import re
    res = re.search(r"\d{4}-\d{1,2}-\d{1,2}", string)
    if res is not None:
        y, m, d = res.group(0).split("-")
        return dt.datetime(int(y), int(m), int(d)).date()
    res = re.search(r"\d{4}/\d{1,2}/\d{1,2}", string)
    if res is not None:
        y, m, d = res.group(0).split("/")
        return dt.datetime(int(y), int(m), int(d)).date()
    res = re.search(r"\d{4}[0,1]{1}\d{1}\d{2}", string)
    if res is not None:
        y, m, d = res.group(0)[:4], res.group(0)[4:6], res.group(0)[6:]
        return dt.datetime(int(y), int(m), int(d)).date()
    res = re.search(r"\d{4}年\d{1,2}月\d{1,2}日", string)
    if res is not None:
        res = res.group(0)
        res = res.replace("年", "-")
        res = res.replace("月", "-")
        res = res.replace("日", "")
        y, m, d = res.split("-")
        return dt.datetime(int(y), int(m), int(d)).date()
    return None


def table2kvalue(table, index_name, column_name, value_name):
    """用于转换DataFrame形式的数据至key-value.

    parameters:
    -----------
    table -- DataFrame
    index_name -- 索引类别名称
    column_name -- 列类别名称
    value_name -- 值名称

    returns:
    --------
    pd.DataFrame
    """

    df = pd.DataFrame()
    table.index.name = index_name
    for col in table.columns:
        c = table[[col]].reset_index().rename(columns={col: value_name})
        c[column_name] = col
        df = pd.concat([df, c], ignore_index=True)

    return df


if __name__ == '__main__':

    pass



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from datetime import timedelta
from scipy.optimize import minimize
import os

import fund_alchemy.general_tools as gt
import fund_alchemy.figure as fg
from fund_alchemy.custodian_table import standard_custodian_table

def generate_review_part_one(df, start_date, end_date, output_path):
    """生成第一部分内容.

    parameters:
    -----------
    df -- pd.DataFrame
        两列净值序列，index为datetime.date，第一列为业绩基准，第二列为产品净值
    start_date -- str
    end_date -- str

    returns:
    ----------
    生成第一部分净值分析内容
    """
    
    fg.draw_overview(df, fre="d")
    plt.savefig(output_path+'净值分析.svg', dpi=300)
    plt.close()
    
    #生成表2
    temp = pd.DataFrame(fg.comprehensive_analysis(df[df.columns[0]], rf=0.03), index =['成立以来-%s'%df.columns[0]])
    temp = pd.concat([temp,pd.DataFrame(fg.comprehensive_analysis(df[df.columns[1]], rf=0.03),
                                        index =['成立以来-%s'%df.columns[1]])])
    temp = pd.concat([temp,pd.DataFrame(fg.comprehensive_analysis(df[df.columns[0]][df.index[-1]-timedelta(364):], rf=0.03),
                                        index =['近一年-%s'%df.columns[0]])])
    temp = pd.concat([temp,pd.DataFrame(fg.comprehensive_analysis(df[df.columns[1]][df.index[-1]-timedelta(364):], rf=0.03),
                                        index =['近一年-%s'%df.columns[1]])])
    temp = pd.concat([temp,pd.DataFrame(fg.comprehensive_analysis(df[df.columns[0]][df.index[-1]-timedelta(182):], rf=0.03),
                                        index =['近六个月-%s'%df.columns[0]])])
    temp = pd.concat([temp,pd.DataFrame(fg.comprehensive_analysis(df[df.columns[1]][df.index[-1]-timedelta(182):], rf=0.03),
                                        index =['近六个月-%s'%df.columns[1]])])
    temp = pd.concat([temp,pd.DataFrame(fg.comprehensive_analysis(df[df.columns[0]][df.index[-1]-timedelta(91):], rf=0.03),
                                        index =['近三个月-%s'%df.columns[0]])])
    temp = pd.concat([temp,pd.DataFrame(fg.comprehensive_analysis(df[df.columns[1]][df.index[-1]-timedelta(91):], rf=0.03),
                                        index =['近三个月-%s'%df.columns[1]])])
    temp = temp[['年化收益率','年化波动率','夏普比率','最大回撤','卡玛比率']]
    temp['年化收益率'] = temp['年化收益率'] * 100
    temp['年化波动率'] = temp['年化波动率'] * 100
    temp['最大回撤'] = temp['最大回撤'] * 100
    temp['夏普比率'] = round(temp['夏普比率'] ,2)
    temp['卡玛比率'] = round(temp['卡玛比率'] ,2)
    temp.to_excel(output_path+'净值指标.xlsx')
        
    #文字描述
    value = df[df.columns[0]][gt.d(start_date): gt.d(end_date)]
    r = value/value.shift(1)-1 
    r = r.drop(gt.d(start_date))
    净值变动 = round(value[-1]/value[0]-1,4)
    单日最大盈利 = round(max(r),4)
    单日最大亏损 = round(min(r),4)
    日收益率均值 = round(np.mean(r),4)
    标准差 = round(np.std(r),4)
    
    return {
        '净值变动': 净值变动,
        '单日最大盈利': 单日最大盈利,
        '单日最大亏损': 单日最大亏损,
        '日收益率均值': 日收益率均值,
        '标准差': 标准差,}


def concat_data(input_path, output_path):
    """从净值表获取当期持仓，市价，成本和市值情况的Dataframe.
    需包含起始日前一天的估值表

    parameters:
    -----------
    input_path -- str
    output_path -- str
    config -- pd.ExcelFile
            配置文件
    
    returns:
    --------
    pd.DataFrame
    """

    def daily_data(path, file_name):
        
        custodian = '中信期货'
        filepath = path + '\\' + file_name
        a = standard_custodian_table(filepath, custodian=None)
        raw_data = a.data
        fund_data = raw_data['main_body']
        
        return fund_data
    
    # 持仓、市价、成本统计
    #取区间前一天开始的估值表

    os.chdir(input_path)
    allxls = os.listdir() 

    amount_mat = pd.DataFrame()
    close_mat = pd.DataFrame()
    cost_mat = pd.DataFrame()
    mv_mat = pd.DataFrame()
    df_name = pd.DataFrame()
    for i in allxls:
        d = daily_data(input_path, i)
        d.index = d['name']
        d = d[d.asset_class == '基金']
        d_sum = d.groupby(d.index).sum()
        
        d_amount = d_sum['amount']
        d_close = d_sum['close']
        d_cost = d_sum['cost']
        d_mv = d_sum['mv']
        
        d_amount.name = i.split('_')[-1].split('.')[0]
        d_close.name = i.split('_')[-1].split('.')[0]
        d_cost.name = i.split('_')[-1].split('.')[0]
        d_mv.name = i.split('_')[-1].split('.')[0]
        
        amount_mat = pd.concat([amount_mat,d_amount], axis=1)
        close_mat = pd.concat([close_mat,d_close], axis=1)
        cost_mat = pd.concat([cost_mat,d_cost], axis=1)
        mv_mat = pd.concat([mv_mat,d_mv], axis=1)
        
        d_name = d[['account_code','name']].copy()
        d_name.account_code = d_name.account_code.apply(lambda x : str(x[-6:]) + '.OF')
        df_name = pd.concat([df_name,d_name], axis=0)
        
    df_name = df_name.drop_duplicates(['account_code'])
    df_name = df_name.drop('name',axis=1)
    df_name = df_name.reset_index()
    
    df_name.to_excel(output_path + '对应关系.xlsx')    
    
    return df_name, amount_mat, close_mat, cost_mat, mv_mat 


def get_nav_from_db(df_name, nav_db, start_date, end_date):
    """获取分析区间子基金的净值序列.

    parameters:
    -----------
    df_name -- pd.DataFrame
    nav_data -- pd.DataFrame
    start_date -- '2021-01-01'
    end_date -- '2021-12-31'

    returns:
    --------
    pd.DataFrame
    """

    nav_db_new = nav_db[['基金代码', '日期','单位净值']]
    nav_db_new_mat = pd.pivot_table(nav_db_new, values = '单位净值', index = '日期', columns = '基金代码').T
    nav_db_new_mat = pd.merge(df_name,nav_db_new_mat,how = 'left', left_on= 'account_code', right_on = '基金代码')
    nav_db_new_mat = nav_db_new_mat.drop('account_code',axis=1)
    nav_db_new_mat = nav_db_new_mat.set_index('name')
    nav_db_new_mat = nav_db_new_mat.T
    nav_db_new_mat = nav_db_new_mat.ffill().bfill()
    
    return nav_db_new_mat


def calculate_funds_contributions(df_name, amount_mat, close_mat, cost_mat, mv_mat, div_file, nav_db_new_mat, output_path):
    """获取分析区间各子基金的净值贡献.

    parameters:
    -----------
    df_name -- pd.DataFrame 
    amount_mat -- pd.DataFrame
    close_mat -- pd.DataFrame 
    cost_mat -- pd.DataFrame 
    mv_mat -- pd.DataFrame
    div_file -- pd.DataFrame 
    nav_db_new_mat -- pd.DataFrame

    returns:
    --------
    pd.DataFrame
    """
    
    amount_mat = amount_mat.T
    close_mat = close_mat.T
    cost_mat = cost_mat.T
    mv_mat =  mv_mat.T
    amount_mat.index = pd.to_datetime(amount_mat.index)
    close_mat.index = pd.to_datetime(close_mat.index)
    mv_mat.index = pd.to_datetime(mv_mat.index)

    #投资金额变动    
    amount_mat = amount_mat.fillna(0)
    amount_diff = amount_mat - amount_mat.shift(1)
    #amount_diff = amount_diff.replace(0, np.nan) # 找到申赎的交易日
    close_adj = nav_db_new_mat.shift(1) # 使用前一天的收盘价
    cost_diff = amount_diff * close_adj # 每次申赎的投入资本变动
    diff_cost = np.sum(cost_diff) # 所有申赎的成本变动累加

    #当期红利
    div = div_file.T
    div = div.reset_index()
    div_wind_mat = pd.merge(df_name,div,how = 'left', left_on= 'account_code', right_on = 'index')
    div_wind_mat = div_wind_mat.drop(['account_code','index'],axis = 1)
    div_wind_mat = div_wind_mat.set_index('name')
    div_wind_mat = div_wind_mat.T
    div_income = div_wind_mat * amount_mat 
    div_income_sum = np.sum(div_income) 

    mv_mat = mv_mat.fillna(0)
    initial_mv = mv_mat.iloc[0] # 初始市值
    final_mv = mv_mat.iloc[-1] # 最终市值
    cum_return = round(final_mv - initial_mv - diff_cost + div_income_sum,2)
    cum_return.name = '累计收益'
    
    contribution = round(100 * cum_return / np.abs(cum_return.sum()),2)
    contribution.name = '贡献度（%）'
    
    con_mat = pd.concat([cum_return, contribution], axis=1)
    con_mat_sort = con_mat.sort_values(by='贡献度（%）', ascending=False)
    con_mat_sort.to_excel(output_path+'贡献度.xlsx')
    
    return con_mat_sort


def fund_indicators(fund_list, nav_db):
    """获取分析区间各子基金的净值贡献.

    parameters:
    -----------
    df_name -- pd.DataFrame 
    amount_mat -- pd.DataFrame
    close_mat -- pd.DataFrame 
    cost_mat -- pd.DataFrame 
    mv_mat -- pd.DataFrame
    div_file -- pd.DataFrame 
    nav_db_new_mat -- pd.DataFrame

    returns:
    --------
    pd.DataFrame
    """
    
    metrics_mat_成立以来 = pd.DataFrame()
    metrics_mat_近一年 = pd.DataFrame()
    metrics_mat_近半年 = pd.DataFrame()
    
    for fund in fund_list:
        data = nav_db[nav_db.基金代码 == fund] 
        nav = data['单位净值'].astype('float')
        nav.index = data.日期     
        
        temp = nav.copy()#成立以来
        temp.name = data['基金代码'][0]
        metrics = fg.comprehensive_analysis(temp)
        nav_metrics = pd.DataFrame([metrics])
        metrics_mat_成立以来 = pd.concat([metrics_mat_成立以来,nav_metrics],ignore_index=True)        
               
        temp = nav[nav.index[-1]-timedelta(364):].copy()#近一年
        temp.name = data['基金代码'][0]
        metrics = fg.comprehensive_analysis(temp)
        nav_metrics = pd.DataFrame([metrics])
        metrics_mat_近一年 = pd.concat([metrics_mat_近一年,nav_metrics],ignore_index=True)

        temp = nav[nav.index[-1]-timedelta(182):].copy()#近半年
        temp.name = data['基金代码'][0]
        metrics = fg.comprehensive_analysis(temp)
        nav_metrics = pd.DataFrame([metrics])
        metrics_mat_近半年 = pd.concat([metrics_mat_近半年,nav_metrics],ignore_index=True)   
        
    column_name = ['基金名称'] + list(metrics_mat_成立以来.columns)
    metrics_mat_成立以来['基金名称'] = name_list
    metrics_mat_近一年['基金名称'] = name_list
    metrics_mat_近半年['基金名称'] = name_list
    
    metrics_mat_成立以来 = metrics_mat_成立以来.loc[:,column_name]
    metrics_mat_近一年 = metrics_mat_近一年.loc[:,column_name]
    metrics_mat_近半年 = metrics_mat_近半年.loc[:,column_name]
    
    fund_list = fund_list.rename('基金代码')
    metrics_mat_成立以来.index = fund_list
    metrics_mat_近一年.index = fund_list
    metrics_mat_近半年.index = fund_list
    
    writer = pd.ExcelWriter('附录2_子基金阶段性业绩表现.xlsx')
    metrics_mat_成立以来.to_excel(writer, '成立以来')    
    metrics_mat_近一年.to_excel(writer, '近一年') 
    metrics_mat_近半年.to_excel(writer, '近半年')
    writer.save()
    
    return 0


def regression_with_industry_index(Xs, y, citic_map, output_path):

    fund_nav = pd.DataFrame(y)
    fund_nav.columns =['nav']
    reg_data = fund_nav.join(Xs).copy()
    reg_data = reg_data.astype(float).pct_change().dropna()
    reg_data['cons'] = 1
    reg_data = reg_data.rename(columns=citic_map)

    Y = 'nav'
    X = [x for x in reg_data.columns if x != Y]
    lb = [0 for i in range(len(reg_data.columns)-1)]
    ub = [1 for i in range(len(reg_data.columns)-1)]
    reg = minimize(
        fun=lambda x: ((reg_data[X].dot(x)-reg_data[Y]
                        ) ** 2).sum(),
        x0=[0 for i in range(len(X))],
        method='SLSQP',
        constraints=(
            {'type': 'ineq', 'fun': lambda x: 1.0 - x.sum()}
            ),
        bounds=list(zip(lb, ub)),
        tol=0.00000000000001)
    reg_params = pd.Series(list(reg.x.round(decimals=6)), index=X)
    reg_params = reg_params.drop('cons').sort_values(ascending=False)
    reg_params.to_excel(output_path + '净值归因.xlsx')
    
    return reg_params      
        

if __name__ == '__main__':
    pass